templates = { }

-- Common stuff for most build() functions
local objectMetatable = {
	basicBuild = function(self, x, y, r)
		self.body = love.physics.newBody(world, x, y, "dynamic")
		
		self.body:setAngle(r)
		
		self.body:setLinearDamping(5)
		self.body:setAngularDamping(5)
		
		self.draw = self.blueprint.draw
		
		self.fixtures = { }
	end,
	
	addFixture = function(self, shape, density)
		local fixture = love.physics.newFixture(self.body, shape, density)
		fixture:setUserData(self)
		
		table.insert(self.fixtures, fixture)
		
		return fixture
	end,
	
	rebuild = function(self, x, y, r)
		--self.body:destroy()
		
		-- this one will presumably be built with a new 'world' variable
		self:basicBuild(x, y, r)
		
		for i, pShape in ipairs(self.blueprint.physicalShapes) do
			self:addFixture(pShape)
		end
		
		if self.blueprint.setupPhysics then
			self.blueprint.setupPhysics(self)
		end
	end,
	
	unmount = function(self)
		-- Do nothing... if something needs to be done, override this
	end,
	
	-- Wrappers
	getPosition = function(self)
		if self.mountedOn then
			return self.mountedOn:getPosition()
		else
			return self.body:getPosition()
		end
	end,
}

objectMetatable.__index = objectMetatable

local function newObject(blueprint)
	local o = { blueprint = blueprint, connectedObjects = { } }
	return setmetatable(o, objectMetatable)
end

local function newBlueprint(blueprint)
	function blueprint.build(x, y, r)
		local o = newObject(blueprint)
		
		if not blueprint.physicalShapes then
			blueprint.physicalShapes = { }
			for i, shape in ipairs(blueprint.shapes) do
				local pShape
				
				if shape.shapeType == "rectangle" then
					pShape = love.physics.newRectangleShape(unpack(shape))
				elseif shape.shapeType == "polygon" then
					pShape = love.physics.newPolygonShape(unpack(shape))
				elseif shape.shapeType == "circle" then
					pShape = love.physics.newCircleShape(unpack(shape))
				else
					print("Unknown shape type: " .. shape.shapeType)
				end
				
				table.insert(blueprint.physicalShapes, pShape)
			end
		end
		
		o:basicBuild(x, y, r)
		
		for i, pShape in ipairs(blueprint.physicalShapes) do
			o:addFixture(pShape)
		end
		
		if blueprint.setupPhysics then
			blueprint.setupPhysics(o)
		end
		
		if blueprint.setupState then
			blueprint.setupState(o)
		end
		
		if blueprint.setupCosmetics then
			blueprint.setupCosmetics(o)
		end
		
		return o
	end
	
	return blueprint
end


local function rectangle(t)
	t.shapeType = "rectangle"
	return t
end

local function polygon(t)
	t.shapeType = "polygon"
	return t
end

local function circle(t)
	t.shapeType = "circle"
	return t
end


-- Player-controlled Vehicle --
-------------------------------

templates.vehicle = newBlueprint { }

templates.vehicle.shapes = {
	rectangle { 20, 20, color = {255, 255, 255} }
}


function templates.vehicle.setupCosmetics(v)
	-- Particle systems for the vehicle's wheels/tracks
	local function makeDirt( )
		local dirt = love.graphics.newParticleSystem(bluePixel, 100)
		
		dirt:setEmissionRate(20)
		dirt:setSpread(math.pi / 4)
		dirt:setSpeed(0)
		dirt:setSizes(2, 2, 0)
		dirt:setParticleLife(2, 4)
		dirt:start()
		
		return dirt
	end
	
	v.dirt1 = v.dirt1 or makeDirt()
	v.dirt2 = v.dirt2 or makeDirt()
	
	return v
end

function templates.vehicle.destroy(self)
	self.dirt1 = nil
	self.dirt2 = nil
	self.body:destroy()
end

function templates.vehicle.updatePositions(self, dt)
	-- Update dirt particle system positions
	self.dirt1:setPosition(self.body:getWorldPoint(0, 5))
	self.dirt1:setDirection(self.body:getAngle() + math.pi)
	self.dirt1:update(dt)
	
	self.dirt2:setPosition(self.body:getWorldPoint(0, -5))
	self.dirt2:setDirection(self.body:getAngle() + math.pi)
	self.dirt2:update(dt)
	
	-- Turn off if idle
	if self.body:getLinearVelocity() == 0 and self.body:getAngularVelocity() == 0 then
		self.dirt1:stop()
		self.dirt2:stop()
	else
		self.dirt1:start()
		self.dirt2:start()
	end
end

function templates.vehicle.drawParticleSystems(self)
    love.graphics.setColor(255, 255, 255)
    love.graphics.draw(self.dirt1)
    love.graphics.draw(self.dirt2)
end

function templates.vehicle.draw(self)
	love.graphics.setColor(255, 255, 255, 255)
    love.graphics.rectangle("fill", -10, -10, 20, 20)
end

templates.vehicle.sideHardpoints = {
	{ 10, 0, angle = 0 },
	{ 0, 10, angle = math.pi / 2 },
	{ -10, 0, angle = math.pi },
	{ 0, -10, angle = 3 * math.pi / 2 },
}

-- ToDo: This needs to be merged with sideHardpoints
templates.vehicle.topHardpoints = {
	{ 0, 0, angle = 0 },
}

templates.vehicle.stateCallSignature = "hooked<bool>"

function templates.vehicle.loadState(self, hooked)
	self.hooked = hooked
end

function templates.vehicle.saveState(self, futureLoadState)
	futureLoadState(self.hooked)
end


-- Simple Long Object --
------------------------

templates.obj1 = newBlueprint { }

templates.obj1.shapes = {
	rectangle { 10, 60 }
}

function templates.obj1.destroy(self)
	self.body:destroy()
end

function templates.obj1.draw(self)
	love.graphics.setColor(200, 200, 200)
	love.graphics.rectangle("fill", -5, -30, 10, 60)
	love.graphics.setColor(200, 255, 200)
	love.graphics.rectangle("fill", -5, -30, 10, 10)
end

function templates.obj1.drawShadow()
	love.graphics.setColor(255, 255, 255)
	love.graphics.rectangle("line", -5, -30, 10, 60)
	love.graphics.rectangle("line", -5, -30, 10, 10)
end

templates.obj1.snapshotScale = 0.4
templates.obj1.sideHardpoints = {
	{-5, 0, angle = math.pi},
	{5, 0, angle = 0},
	{0, 30, angle = math.pi / 2},
	{0, -30, angle = 3 * math.pi / 2}
}
templates.obj1.connectors = { {0, 30, angle = math.pi / 2}, {5, 0, angle = 0} }


-- Simple Always-on Thruster (with infinite fuel supply for now) --
-------------------------------------------------------------------

templates.thruster1 = newBlueprint { }

templates.thruster1.shapes = {
	rectangle {20, 10},
	polygon {-5, -2,  -5, 2,  -15, 5,  -15, -5},
}

function templates.thruster1.setupCosmetics(o)
	-- Make a particle system
	o.particles = love.graphics.newParticleSystem(bluePixel, 30)
	o.particles:setEmissionRate(30)
	o.particles:setParticleLife(0.5, 1)
	o.particles:setSpread(math.pi / 4)
	o.particles:setSpeed(100)
	o.particles:setSizes(2, 2, 1, 0)
	o.particles:start()
end

function templates.thruster1.updatePhysics(self)
	-- Forward thrust
	local r = self.body:getAngle()
	self.body:applyForce(math.cos(r)*2000, math.sin(r)*2000)
end

function templates.thruster1.updatePositions(self, dt)
	-- Update particle system
	self.particles:setPosition(self.body:getWorldPoint(-10, 0))
	self.particles:setDirection(self.body:getAngle() + math.pi)
	self.particles:update(dt)
end

function templates.thruster1.drawParticleSystems(self)
	love.graphics.draw(self.particles)
end

function templates.thruster1.draw(self, drawWireframe)
	local fill = drawWireframe and "line" or "fill"
	love.graphics.setColor(255, 200, 200, 255)
	love.graphics.rectangle(fill, -10, -5, 20, 10)
	love.graphics.setColor(255, 255, 255, 255)
	love.graphics.polygon(fill,  -5, -2,  -5, 2,  -15, 5,  -15, -5)
end

function templates.thruster1.drawShadow()
	templates.thruster1.draw(nil, true)
end

function templates.thruster1.destroy(self)
	self.particles = nil
	self.body:destroy()
end

function templates.thruster1.damage(self)
	explodeObject(self)
end

templates.thruster1.snapshotScale = 0.8
templates.thruster1.sideHardpoints = { }
templates.thruster1.connectors = {
	{ 10, 0, angle = 0 },
	{ 0, 5, angle = math.pi / 2 },
	{ 0, -5, angle = 3 * math.pi / 2 },
}


-- Grappling Hook --
--------------------

templates.hook = newBlueprint { }

templates.hook.shapes = {
	polygon { -5, -5,  5, 0,  2, 0,  -5, -3 },
	polygon { -5, 5,  5, 0,  2, 0,  -5, 3 },
}

function templates.hook.setupPhysics(o)
	-- it's sort of flying, so set less friction
	o.body:setLinearDamping(2)
	o.body:setAngularDamping(2)
	
	-- it's also sort of fast, so use continuous collision detection for it.
	o.body:setBullet(true)
end

function templates.hook.draw(self)
	love.graphics.setColor(255, 100, 255, 255)
	love.graphics.polygon("fill",  -5, -5,  5, 0,  2, 0,  -5, -3)
	love.graphics.polygon("fill",  -5, 5,  5, 0,  2, 0,  -5, 3)
end

function templates.hook.destroy(self)
	if not self.hooked and self.fromObject then
		self.fromObject.hooked = false
	end
	
	self.body:destroy()
end

function templates.hook.damage(self)
	explodeObject(self)
end

function templates.hook.updatePhysics(self)
	if self.linkLife then
		self.linkLife = self.linkLife - 1
		
		if self.linkLife <= 0 then
			explodeObject(self)
		end
	end
end

function templates.hook.collision(self, obj2)
	if self.hooked or not self.fromObject then
		return
	end
	
	-- Move it slightly forward, to look as if the hook pierced the object
	self.body:setPosition(self.body:getWorldPoint(5, 0))
	
	-- Weld the hook and the other object together.
	local x, y = self.body:getWorldPoint(0, 0)
	local joint = love.physics.newWeldJoint(self.body, obj2.body, x, y, false)
	self.hooked = true
	self.hookedTo = obj2
	self.fromObject.hooked = true
	
	if not obj2.wall then
		obj2:callWhenDestroyed(function(self2)
			self.hookedTo = nil
		end)
	end
	
	-- Create a chain from the hook the the vehicle
	local fromx, fromy = self.body:getWorldPoint(0, 0)
	local tox, toy = self.fromObject.body:getWorldPoint(14, 0)
	
	local links = math.floor(distance(fromx, fromy, tox, toy) / 8)
	
	local oldx, oldy = fromx, fromy
	local oldLink = self
	
	local function interpolate(x1, y1, x2, y2, r)
		local x = x1 * (1-r) + x2 * r
		local y = y1 * (1-r) + y2 * r
		return x, y
	end
	
	for i = 1, links do
		x, y = interpolate(fromx, fromy, tox, toy, i / links)
		local newLink = registerObject(templates.chainlink.build(x, y, 0))
		x, y = newLink.body:getPosition()
		
		-- Maybe consider a rope joint instead?
		joint = love.physics.newRopeJoint(oldLink.body, newLink.body,
			oldx * 20, oldy * 20, x * 20, y * 20, 8, true)
		--joint = love.physics.newDistanceJoint(oldLink.body, newLink.body,
		--	oldx, oldy, x, y, true)
		
		newLink.nextChainLink = oldLink
		
		newLink.nextChainLink:callWhenDestroyed(function (self)
			newLink.nextChainLink = nil
		end)
		
		oldLink = newLink
		oldx, oldy = x, y
	end
	
	oldLink.weldedTo = self.fromObject
	
	oldLink.weldedTo:callWhenDestroyed(function (self)
		oldLink.linkLife = 4
		oldLink.weldedTo = nil
	end)
	
	-- Weld the last link to the vehicle
	joint = love.physics.newWeldJoint(oldLink.body, self.fromObject.body, oldx, oldy, false)
	
	-- ToDo: register the joints
end

templates.hook.stateCallSignature = "hooked<bool> hookedTo<int> fromObject<int>"

function templates.hook.loadState(self, hooked, hookedToId, fromObjectId)
	self.fromObject = objectById[fromObjectId]
	self.hookedTo = objectById[hookedToId]
	self.hooked = hooked
	
	if self.hookedTo then
		local x, y = self.body:getWorldPoint(0, 0)
		local joint = love.physics.newWeldJoint(self.body, self.hookedTo.body, x, y, false)
		
		self.hookedTo:callWhenDestroyed(function (self2)
			self.hookedTo = nil
		end)
	end
	
	if self.fromObject then
		self.fromObject.hooked = hooked
		
		self.fromObject:callWhenDestroyed(function (self2)
			self.fromObject = nil
		end)
	end
	
	if fromObjectId ~= 0 and not self.fromObject then
		print("Invalid object id (" .. fromObjectId .. ") given to hook:loadState()")
	end
	
	if hookedToId ~= 0 and not self.hookedTo then
		print("Invalid object id (" .. hookedToId .. ") given to hook:loadState()")
	end
end

function templates.hook.saveState(self, futureLoadState)
	local hookedToId = self.hookedTo and self.hookedTo.id or 0
	local fromObjectId = self.fromObject and self.fromObject.id or 0
	
	futureLoadState(self.hooked, hookedToId, fromObjectId)
end


-- A link in the grapple-hook's chain --
----------------------------------------

templates.chainlink = newBlueprint { }

templates.chainlink.shapes = {
	circle { 2 }
}

function templates.chainlink.setupPhysics(o)
	-- it's sort of flying, so set less friction
	o.body:setLinearDamping(2)
	o.body:setAngularDamping(2)
	
	-- it's also sort of small, so use continuous collision detection for it.
	o.body:setBullet(true)
end

function templates.chainlink.draw(self)
	love.graphics.setColor(0, 0, 255, 192)
	love.graphics.circle("fill", 0, 0, 2, 6)
	love.graphics.setColor(255, 255, 255, 255)
	love.graphics.circle("line", 0, 0, 2, 6)
end

function templates.chainlink.updatePhysics(self)
	if self.linkLife then
		self.linkLife = self.linkLife - 1
		
		if self.linkLife <= 0 then
			addChainLinkDispersion(self.body:getPosition())
			destroyObject(self)
		end
	end
end

function templates.chainlink.destroy(self)
	if self.weldedTo then
		self.weldedTo.hooked = false
	end
	
	if self.nextChainLink then
		self.nextChainLink.linkLife = 3
	end
	
	self.body:destroy()
end

function templates.chainlink.damage(self)
	explodeObject(self)
end

templates.chainlink.stateCallSignature = "nextlink<int> weldedTo<int> linklife<int>"

-- ToDo: make a common function for 'get id, complain about id, add destroy callback'
function templates.chainlink.loadState(self, nextChainLinkId, weldedToId, linkLife)
	self.nextChainLink = objectById[nextChainLinkId]
	self.weldedTo = objectById[weldedToId]
	self.linkLife = linkLife
	
	if self.linkLife == -1 then
		self.linkLife = nil
	end
	
	if self.nextChainLink then
		local oldx, oldy = self.body:getPosition()
		local x, y = self.nextChainLink.body:getPosition()
		
		joint = love.physics.newRopeJoint(self.body, self.nextChainLink.body,
			oldx * 20, oldy * 20, x * 20, y * 20, 8, true)
		
		self.nextChainLink:callWhenDestroyed(function (self2)
			self.nextChainLink = nil
		end)
	end
	
	if self.weldedTo then
		local x, y = self.body:getPosition()
		love.physics.newWeldJoint(self.body, self.weldedTo.body, x, y, false)
		
		self.weldedTo:callWhenDestroyed(function (self2)
			self.linkLife = 4
			self.weldedTo = nil
		end)
	end
	
	if nextChainLinkId ~= 0 and not self.nextChainLink then
		print("Invalid object id (" .. nextChainLinkId .. ") given to chainlink:loadState()")
		print("Self id:", self.id)
	end
	
	if weldedToId ~= 0 and not self.weldedTo then
		print("Invalid weld object id (" .. weldedToId .. ") given to chainlink:loadState()")
	end
end

function templates.chainlink.saveState(self, futureLoadState)
	local nextChainLinkId = self.nextChainLink and self.nextChainLink.id or 0
	local weldedToId = self.weldedTo and self.weldedTo.id or 0
	
	futureLoadState(nextChainLinkId, weldedToId, self.linkLife or -1)
end


-- Side-mounted grappling hook gun --
-------------------------------------

templates.grapplinghookgun = newBlueprint { }

templates.grapplinghookgun.shapes = {
	rectangle { 16, 16 }
}

function templates.grapplinghookgun.draw(self)
	love.graphics.setColor(64, 96, 64)
	love.graphics.rectangle("fill", -8, -8, 16, 16)
	
	if not self or not self.hooked then
		templates.hook.draw()
	end
end

function templates.grapplinghookgun.drawShadow()
	love.graphics.setColor(192, 192, 192)
	love.graphics.rectangle("line", -8, -8, 16, 16)
	
	templates.hook.draw()
end

function templates.grapplinghookgun.activate(self)
	fireHook(self)
end

templates.grapplinghookgun.connectors = {
	{ -8, 0, angle = math.pi },
	{ 0, 8, angle = math.pi / 2 },
	{ 0, -8, angle = 3 * math.pi / 2 },
}

templates.grapplinghookgun.sideHardpoints = {
	{ -8, 0, angle = math.pi },
	{ 0, 8, angle = math.pi / 2 },
	{ 0, -8, angle = 3 * math.pi / 2 },
}

function templates.grapplinghookgun.destroy(self)
	self.body:destroy()
end

function templates.grapplinghookgun.damage(self)
	explodeObject(self)
end

templates.grapplinghookgun.stateCallSignature = "hooked<bool>"

function templates.grapplinghookgun.loadState(self, hooked)
	self.hooked = hooked
end

function templates.grapplinghookgun.saveState(self, futureLoadState)
	futureLoadState(self.hooked)
end


-- Small continuous laser --
----------------------------

templates.laser1 = newBlueprint { }

templates.laser1.shapes = {
	rectangle { 20, 10 },
	rectangle { 10, 0, 4, 6 },
}

function templates.laser1.setupState(self)
	self.energy = 0
end

function templates.laser1.draw(self)
	love.graphics.setColor(128, 128, 192)
	love.graphics.rectangle("fill", -10, -5, 20, 10)
	love.graphics.setColor(32, 32, 32)
	love.graphics.rectangle("fill", -5, -3, 17, 6)
	love.graphics.setColor(255, 0, 0)
	love.graphics.rectangle("fill", -5, -3, 17 * (self and self.energy or 1), 6)
	love.graphics.setColor(0, 0, 255)
	love.graphics.rectangle("fill", -3, -1, 15, 2)
end

function templates.laser1.drawShadow()
	love.graphics.setColor(128, 128, 192)
	love.graphics.rectangle("line", -10, -5, 20, 10)
	love.graphics.setColor(255, 0, 0)
	love.graphics.rectangle("line", -5, -3, 17, 6)
	love.graphics.setColor(0, 0, 255)
	love.graphics.rectangle("line", -3, -1, 15, 2)
end

templates.laser1.snapshotScale = 0.8
templates.laser1.sideHardpoints = {
	{ 0, 5, angle = math.pi / 2 },
	{ 0, -5, angle = 3 * math.pi / 2 },

}

templates.laser1.connectors = {
	{ -10, 0, angle = math.pi },
	{ 0, 5, angle = math.pi / 2 },
	{ 0, -5, angle = 3 * math.pi / 2 },
}

function templates.laser1.updatePhysics(self, dt)
	if self.laserActivated then
		if self.energy <= 0 then
			self.laserActivated = false
			return
		end
		
		self.energy = self.energy - dt
		--createLaserBeam(self, 9, 1000, nil, true)
		createLaserBeam(self, 9, 1000, 0.1)
	elseif self.energy < 1 then
		self.energy = self.energy + dt / 2
	end
end

function templates.laser1.activate(self)
	if self.energy > 0.5 then
		self.laserActivated = true
	end
end

function templates.laser1.destroy(self)
	self.body:destroy()
end

function templates.laser1.damage(self)
	explodeObject(self)
end

templates.laser1.stateCallSignature = "energy<float>"

function templates.laser1.loadState(self, energy)
	self.energy = minMax(energy, 0, 1)
end

function templates.laser1.saveState(self, futureLoadState)
	futureLoadState(self.energy)
end

function templates.laser1.showInfo(self, infoBox)
	infoBox.showText(string.format("Energy: %deu", minMax(self.energy * 100, 0, 100)))
end


-- Top-mounted torque upgrade module --
---------------------------------------

templates.torqueupgrade = newBlueprint { }

templates.torqueupgrade.build = nil

function templates.torqueupgrade.buildUnmounted()
	return newObject(templates.torqueupgrade)
end

function templates.torqueupgrade.remountOn(self, object, topHardpoint)
	local x, y = unpack(topHardpoint)
	self.shape = love.physics.newRectangleShape(x, y, 10, 10)
	self.fixture = object:addFixture(self.shape, 0.5)
end

function templates.torqueupgrade.mountOn(object, topHardpoint)
	local o = templates.torqueupgrade.buildUnmounted()
	
	templates.torqueupgrade.remountOn(o, object, topHardpoint)
	
	return o
end

function templates.torqueupgrade.draw(self)
	--love.graphics.setColor(32, 32, 32)
	--love.graphics.rectangle("line", -7, -7, 14, 14)
	
	local stencil = love.graphics.newStencil(function()
		love.graphics.setColor(255, 255, 255)
		love.graphics.rectangle("fill", -10, -2, 20, 4)
	end)
	
	-- Watch out for nested stencils?
	love.graphics.setInvertedStencil(stencil)
	love.graphics.setColor(0, 0, 255)
	love.graphics.circle("fill", 0, 0, 6, 12)
	love.graphics.setInvertedStencil()
end

function templates.torqueupgrade.drawShadow()
	love.graphics.setColor(32, 32, 32)
	love.graphics.rectangle("line", -7, -7, 14, 14)
	love.graphics.setColor(0, 0, 255)
	love.graphics.circle("line", 0, 0, 5, 12)
end

function templates.torqueupgrade.destroy(self)
	self:unmount()
	self.fixture:destroy()
end


----

-- Some template housework
for name in pairs(templates) do
	templates[name].name = name
end
