local socket = require("socket")

local connectTo, connectPort = settings.clientAddress, settings.clientPort

messageSignatures = {
	-- Tick reports for desynchronization
	["tickReport"] = "tick<int> x<float> y<float> r<float>",
	
	-- Keyboard movement
	forward = "", backward = "", left = "", right = "",
	forwardStop = "", backwardStop = "", leftStop = "", rightStop = "",
	
	-- Mouse movement
	["guideAngle"] = "r<float>",
	["stopGuiding"] = "",
	
	-- Actions
	["fireLaser"] = "",
	["fireVehicleHook"] = "",
	["radialSignal"] = "",
	["materializeObject"] = "type<string> x<float> y<float> r<float>",
	["materializeAndConnectObject"] = "type<string> to<int> hardpoint<int> connector<int>",
	["mountObject"] = "type<string> obj<int> topHp<int>",
	["disintegrateObject"] = "id<int>",
	
	-- Blueprint creation
	["addConnectedObjectToBlueprint"] = "type<string> toObjId<int> hardpointNr<int> connectorNr<int>",
	["addObjectToBlueprint"] = "type<string> x<float> y<float> r<float>",
	["materializeCustomBlueprint"] = "x<float> y<float> r<float>",
}

-- Send a message with either of:
--   sendToServer("guideAngle", {x = 10, y = 13})
--   sendToServer("guideAngle", 10, 13)
--   sendCall.guideAngle(10, 13)



local client = {}
local partialMsg = ""


function startClient()
	local s, err = socket.connect(connectTo, connectPort)
	
	client.socket = s
	
	if not client.socket then
		print("Connection failed (" .. err .. ")")
		return
	end
	
	client.socket:settimeout(0)
	client.socket:setoption("tcp-nodelay", true)
end

function stopClient()
	client.socket:close()
	client.socket = nil
	print("Derp? Connection lost..")
end

function connected()
	return client.socket ~= nil
end

function checkClient()
	while true do
		if not client.socket then
			return
		end
		
		local msg, err, pMsg = client.socket:receive("*l")
		
		if pMsg then
			partialMsg = partialMsg .. pMsg
		end
		
		if err == "timeout" then
			return
		end
		
		if err == "closed" then
			stopClient()
		elseif err then
			io.stderr.write("[client - socket:receive()] - " .. err .. "\n")
			stopClient()
		elseif msg then
			handleServerMessage(partialMsg .. msg)
			partialMsg = ""
		end
	end
end

function sendToServerRaw(msg)
	if not client.socket then
		return
	end
	
	client.socket:send(msg .. "\n")
end


tickMessages = { }
newMessages = { }

currentTick = 1
nextTick = 1

estimatedServerTime = nil

-- We need to somehow get this from the server
local serverTimeStep = 1 / 60

-- Create a buffer of messages for this many seconds:
--latency = 0.02
latency = 0.02

function handleServerMessage(msg)
	-- Advance tick
	if msg == ">" then
		
		-- Should we delay processing a bit? If latency is enabled and high enough,
		-- this should help smooth the physics update rate a bit, when the receive
		-- rate of network packets is chaotic
		
		local now = love.timer.getMicroTime()
		
		if not estimatedServerTime then
			estimatedServerTime = now
		else
			estimatedServerTime = estimatedServerTime + serverTimeStep
		end
		
		if estimatedServerTime > now then
			estimatedServerTime = now
		end
		
		local lagSpike = now - estimatedServerTime
		if not maxLagSpike or lagSpike > maxLagSpike then
			maxLagSpike = lagSpike
		end
		
		if nextTick % (60 * 3) == 0 then
			showMaxLagSpike = maxLagSpike
			maxLagSpike = nil
		end
		
		newMessages.processAt = estimatedServerTime + latency
		
		tickMessages[nextTick] = newMessages
		newMessages = { }
		
		nextTick = nextTick + 1
	else
		-- Handle ping replies instantly, don't delay processing.
		local tick = string.match(msg, "^server: pingReply (%d+)$")
		if tick then
			tickPingReply(tonumber(tick))
		else
			-- Normal message... process it later.
			table.insert(newMessages, msg)
		end
	end
end

function processTickMessages()
	if tickMessages[currentTick] then
		if tickMessages[currentTick].processAt > love.timer.getMicroTime() then
			return false
		end
		
		for i, msg in ipairs(tickMessages[currentTick]) do
			parseMessage(msg)
		end
		
		tickMessages[currentTick] = nil
		currentTick = currentTick + 1
		
		return true
	else
		return false
	end
end

players = { }
myName = "spectator"

tickReports = { }
newestTickReport = nil

function parseMessage(msg)
	local player, callStart = string.match(msg, "^player<([^>]+)>: ()")
	local server = false
	
	if not player then
		server, callStart = string.match(msg, "^(server): ()")
	end
	
	if player == "spectator" then
		return
	end
	
	msg = string.sub(msg, callStart or 1)
	
	-- Handle special player-less cases
	if string.find(msg, "^state%-broadcast") then
		-- Game state resync (this one can be sent by a player too)
		print("State broadcast received; reloading state.")
		loadGameState(msg)
		
		return
	end
	
	if server then
		local newName = string.match(msg, "^Welcome! You are: name<([^>]+)>$")
		if newName then
			-- We just connected and received a name to act by
			myName = newName
			sendToServerRaw("create-player-vehicle")
		end
		
		local lostName = string.match(msg, "^player%-disconnected name<([^>]+)>$")
		if lostName and players[lostName] then
			-- Someone else disconnected
			explodeObject(players[lostName])
			players[lostName] = nil
		end
		
		-- Reporting of desynchronization level
		local repTick = string.match(msg, "tick (%d+) report request")
		if repTick then
			sendTickReport(tonumber(repTick))
		end
		
		return
	end
	
	if player and not players[player] then
		-- Someone else connected and is requesting to join
		if msg == "create-player-vehicle" then
			players[player] = makePlayerVehicle(player)
		end
		
		return
	end
	
	if not player then
		print("Huh? Received a message from nobody: " .. msg)
		return
	end
	
	-- Handle the rest as broadcasted function calls
	-- ToDo: the debug stacktrace shows the function as 'f'... figure out a way
	--  to make it show the original function's name instead
	local f, args = unmarshalCall(msg)
	if f and args then
		f(players[player], unpack(args))
	end
end

function tickPingReply(tick)
	local reports = tickReports[tick]
	
	if not reports or not reports[myName] or not reports[myName].time then
		print("Got a tick reply for an inexistent tick (" .. tick .. ")... " ..
				"maybe it is too old?")
		return
	end
	
	ping = love.timer.getTime() - reports[myName].time
end

function tickReport(fromVehicle, tick, x, y, r)
	local player = fromVehicle.player
	local reports = tickReports[tick]
	
	if not reports or not reports[player] then
		return
	end
	
	if player == myName then
		local sentAt = reports[player].time
		local receivedAt = love.timer.getTime()
		
		-- How long it took between sending this command, getting it back,
		-- and then processing it. This includes all possible latencies
		-- (network, server tick interval, network again, and user-enforced
		-- processing delay latency)
		effectiveCommandLatency = receivedAt - sentAt
	else
		-- Get the values from our simulation
		local sr, sx, sy = unpack(reports[player])
		
		-- Make them undergo the same number->text->number transformation
		sr, sx, sy = tonumber(sr .. ""), tonumber(sx .. ""), tonumber(sy .. "")
		
		-- And compare with the reported ones
		reports[player].diff = math.abs(x - sx) + math.abs(y - sy) + math.abs(r - sr)
	end
	
	newestTickReport = reports
end

function sendTickReport(tick)
	-- Save player positions for this tick
	tickReports[tick] = { }
	for player, vehicle in pairs(players) do
		tickReports[tick][player] = {vehicle.body:getAngle(), -- one value
		                             vehicle.body:getPosition()} -- two values
	end
	
	if objects.vehicle and myName ~= "spectator" then
		local x, y = objects.vehicle.body:getPosition()
		local r = objects.vehicle.body:getAngle()
		
		-- Send a ping to the server for this tick... ping replies will be
		-- handled without delay, so this should be sent before the report.
		sendToServerRaw("ping " .. tick)
		
		-- Send our coordinates so that other people can compare them with those from
		-- their simulation
		sendToServer("tickReport", tick, x, y, r)
		
		-- Save the current time so we can see the latency when we receive it again
		tickReports[tick][myName] = {time = love.timer.getTime()}
	end
	
	-- Delete very old reports
	for k, v in pairs(tickReports) do
		if k < tick - 1000 then
			tickReports[k] = nil
		end
	end
end


local function generateObjectStateMessageType(object)
	if object.blueprint.stateCallSignature and not object.blueprint.stateMsgType then
		local msgType = generateTypeFromSignature("state",
				object.blueprint.stateCallSignature)
		msgType.func = object.blueprint.loadState
		object.blueprint.stateMsgType = msgType
	end
end

function saveGameState( )
	local msg = ""
	
	for object in pairs(dynamicObjects) do
		local x, y, r, vx, vy, vr = 0, 0, 0, 0, 0, 0
		
		if object.body then
			x, y = object.body:getPosition()
			r = object.body:getAngle()
			
			vx, vy = object.body:getLinearVelocity()
			vr = object.body:getAngularVelocity()
		end
		
		generateObjectStateMessageType(object)
		
		-- Object-specific state
		local objStateMsg
		if object.blueprint.stateMsgType then
			local function futureLoadState(...)
				objStateMsg = marshalCall(object.blueprint.stateMsgType, {...}, false)
			end
			object.blueprint.saveState(object, futureLoadState)
		end
		objStateMsg = objStateMsg or ""
		
		local cto, ccon, chp = 0, 0, 0
		if object.connectedTo then
			cto = object.connectedTo.toObject.id
			ccon = object.connectedTo.connectorNr
			chp = object.connectedTo.hardpointNr
			
			-- Top-mounted instead of side-welded? Then negate the hardpoint id.
			if object.mountedOn then
				chp = -chp
			end
		end
		
		msg = msg .. "object - id<" .. object.id .. "> " ..
		             "type<" .. object.blueprint.name .. "> " ..
		             "name<" .. (object.player or "n/a") .. "> " ..
		             "pos<" .. x .. "," .. y .. "," .. r .. "> " ..
		             "velocity<" .. vx .. "," .. vy .. "," .. vr .. "> " ..
		             "cto<".. cto .. "> ccon<".. ccon .. "> chp<".. chp .. "> " ..
		             "state{" .. objStateMsg .. "} "
	end
	
	return msg
end

function loadGameState(msg)
	local objectPattern = "object %- id<(%d+)> type<([^>]+)> name<([^>]+)> " ..
	                      "pos<([%d%.e-]+),([%d%.e-]+),([%d%.e-]+)> " ..
	                      "velocity<([%d%.e-]+),([%d%.e-]+),([%d%.e-]+)> " ..
	                      "cto<(%d+)> ccon<(%d+)> chp<(-?%d+)> " ..
	                      "state{([^}]*)}"
	
	oldObjectById = objectById
	
	dynamicObjects = { }
	objectById = { }
	nextId = 1
	
	-- Be brave... reset everything!
	buildNewWorld()
	
	local objStates = { }
	local connectObjects = { }
	
	-- ToDo: this obviously needs validation
	for id, type, name, x, y, r, vx, vy, vr, cto, ccon, chp, state
			in string.gmatch(msg, objectPattern) do
		-- Random comment to separate these two weird blocks of code
		id, x, y, r, vx, vy, vr, cto, ccon, chp = tonumber(id), tonumber(x),
			tonumber(y), tonumber(r), tonumber(vx), tonumber(vy), tonumber(vr),
			tonumber(cto), tonumber(ccon), tonumber(chp)
		
		local object = oldObjectById[id]
		
		if not templates[type].mountOn then
			if not object then
				object = templates[type].build(x, y, r)
			else
				object:rebuild(x, y, r)
			end
			
			object.body:setLinearVelocity(vx, vy)
			object.body:setAngularVelocity(vr)
		else
			if not object then
				object = templates[type].buildUnmounted()
			else
				object:unmount()
			end
		end
		
		registerObject(object, id)
		
		object.connectedObjects = { }
		
		if cto ~= 0 then
			table.insert(connectObjects, {obj = object, cto = cto,
					chp = chp, ccon = ccon})
		else
			object.connectedTo = nil
			object.mountedOn = nil
		end
		
		if type == "vehicle" then
			players[name] = object
			object.player = name
			
			if name == myName then
				objects.vehicle = object
			end
		end
		
		local params = string.match(state, "^state ?(.*)$")
		if params and object.blueprint.stateCallSignature then
			table.insert(objStates, {obj = object, stateMsg = params})
		end
	end
	
	-- Object connections depend on all objects to exist first
	for i, c in ipairs(connectObjects) do
		if not objectById[c.cto] then
			print("Attempting to connect to unknown object id (" .. c.cto .. ") " ..
				"while loading state information.")
		else
			if c.chp > 0 then
				-- Side-welded
				-- Move it temporarily so it's in the proper position while the
				-- weld joint is created.
				local x, y = c.obj.body:getPosition()
				local r = c.obj.body:getAngle()
				connectObjectTo(c.obj, objectById[c.cto], c.chp, c.ccon)
				c.obj.body:setPosition(x, y)
				c.obj.body:setAngle(r)
			else
				-- Top-mounted
				connectObjectOn(c.obj, objectById[c.cto], -c.chp)
			end
		end
	end
	
	-- Same for object states
	for i, s in ipairs(objStates) do
		local object, params = s.obj, s.stateMsg
		
		generateObjectStateMessageType(object)
		local msgType = object.blueprint.stateMsgType
		
		local p = unmarshalCallParams(msgType, params)
		if p then
			object.blueprint.loadState(object, unpack(p))
		end
	end
	
	guiDesyncNotice = 0.5
end
