-- A signature of this form:
--   messageSignatures.guideAngle = "x<int> y<int>"
-- will be used to generate:
--   messageTypes.guideAngle.arguments = { "x", "y" }
--   messageTypes.guideAngle.types = { "int", "int" }
--   messageTypes.guideAngle.func = _G["guideAngle"] -- looked up in love.load()
--   messageTypes.guideANgle.pattern = "x<(-?%d+)> y<(-?%d+)>"

-- Send a message with either of:
--   sendToServer("guideAngle", {x = 10, y = 13})
--   sendToServer("guideAngle", 10, 13)
--   sendCall.guideAngle(10, 13)


-- No point in validating on send.


-- Transform a list of parameters into a string of the form:
--   paramName1<value1> paramName2<value2> ...
-- (parameter names and types taken from the message signature defined in client.lua)
function marshalCall(msgType, params, paramsAreNamed)
	--local msgType = messageTypes[messageName]
	local msg = msgType.name
	
	if paramsAreNamed then
		-- Call style: sendToServer("guideAngle", {x = 10, y = 13})
		for i, prm in ipairs(msgType.arguments) do
			local pString = params[prm]
			if not pString then
				print("Warning: no parameter given for argument '" .. prm .. "'" ..
				      " of function '" .. msgType.name .. "'")
				pString = ""
			end
			msg = msg .. " " .. prm .. "<" .. pString .. ">"
		end
	else
		-- Call style: sendToServer("guideAngle", 10, 13)
		for i, arg in ipairs(msgType.arguments) do
			local pString = params[i]
			
			if msgType.types[i] == "bool" then
				pString = pString and "t" or "f"
			elseif not pString then
				print("Warning: no parameter given for argument '" .. arg .. "'" ..
				      " of function '" .. msgType.name .. "'")
				pString = ""
			end
			
			msg = msg .. " " .. arg .. "<" .. pString .. ">"
		end
	end
	
	return msg
end

-- Fetch a message type from the ones generated from client.lua's signatures and use it
function sendToServer(messageName, ...)
	local params = { ... }
	local msgType = messageTypes[messageName]
	
	if not msgType then
		error("Message signature not defined for: '" .. messageName .. "'")
	end
	
	if type(params[1]) == "table" then
		local msg = marshalCall(msgType, params[1], true)
		sendToServerRaw(msg)
	else
		local msg = marshalCall(msgType, params, false)
		sendToServerRaw(msg)
	end
end

-- Syntactic suger for the above function, to allow calls of the type:
--  sendCall.guideAngle(10, 20)
-- Or is it semantic sugar?
sendCallMetatable = {
	__index = function(t, key)
		-- Return a closure
		return function (...)
			sendToServer(key, ...)
		end
	end
}

sendCall = setmetatable({}, sendCallMetatable)


local varTypePatterns = {
	["<int>"] = "<(-?%d+)>",
	["<float>"] = "<([%d%.e-]+)>",
	["<string>"] = "<([^>]*)>",
	["<bool>"] = "<([tf])>",
}

-- The signatures are defined in client.lua
function generateMessageTypes()
	messageTypes = { }
	
	for name, signature in pairs(messageSignatures) do
		messageTypes[name] = generateTypeFromSignature(name, signature)
		messageTypes[name].func = _G[name]
		
		if not messageTypes[name].func then
			error("Incorrect message signature; no function named '" ..
					name .. "' exists.")
		end
	end
end

function generateTypeFromSignature(name, signature)
	local type = { }
	
	type.name = name
	type.arguments = {}
	type.types = {}
	
	for varName, varType in string.gmatch(signature, "([%a_]+)<([%a_]+)>") do
		table.insert(type.arguments, varName)
		table.insert(type.types, varType)
	end
	
	type.pattern = string.gsub(signature, "<[%a_]+>", varTypePatterns)
	
	return type
end

-- Returns a function and a table with parameters to be given to it.
function unmarshalCall(message)
	local messageName, paramsStart = string.match(message, "^([%a_]+) ?()")
	
	-- Make sure a signature was defined for this message
	if not messageName then
		io.stderr:write("[client - unmarshalCall()] - malformed message: '" ..
				message .. "'\n")
		return
	end
	
	local msgType = messageTypes[messageName]
	local paramsString = string.sub(message, paramsStart)
	
	return msgType.func, unmarshalCallParams(msgType, paramsString)
end

function unmarshalCallParams(msgType, paramsString)
	if not msgType then
		io.stderr:write("[client - unmarshalCall()] - unknown message type " ..
				"in message: '" .. paramsString .. "'\n")
		return
	end
	
	local params
	
	if paramsString == "" then
		params = { }
	else
		params = { string.match(paramsString, msgType.pattern) }
	end
	
	-- Make sure all required parameters exist and are valid
	for i, argType in ipairs(msgType.types) do
		if argType == "int" or argType == "float" then
			params[i] = tonumber(params[i])
			if not params[i] then
--				io.stderr:write("[client - unmarshalCall()] - expected integer " ..
--						"for argument '" .. msgType.arguments[i] .. "' " ..
--						"in message: '" .. message .. "'\n")
				-- ToDo: Can't use the above error, since string.match is all-or-none
				io.stderr:write("[client - unmarshalCall()] - invalid parameters " ..
						"for function '" .. msgType.name .. "' " ..
						"in message: '" .. paramsString .. "'\n")
				return
			end
		elseif argType == "bool" then
			params[i] = (params[i] == "t") and true or false
		end
	end
	
	-- Not a security issue, but... a good warning to help debug, I think...
	if #params > #msgType.types then
		io.stderr:write("[client - unmarshalCall()] - too many parameters " ..
				"given for the function '" .. msgType.name .. "' " ..
				"in message: '" .. paramsString .. "'\n")
		return
	end
	
	return params
end

